<?php declare(strict_types=1);

namespace Res\ResChannableConnector\Core\Api\Controller;

use Doctrine\DBAL\Connection;
use Exception;
use Res\ResChannableConnector\Setting\Service\Config\ResChannableConfigService;
use Shopware\Core\Checkout\Cart\Delivery\Struct\DeliveryDate;
use Shopware\Core\Checkout\Cart\Delivery\Struct\DeliveryTime;
use Shopware\Core\Checkout\Cart\Tax\Struct\CalculatedTax;
use Shopware\Core\Checkout\Cart\Tax\Struct\TaxRule;
use Shopware\Core\Checkout\Cart\Tax\Struct\TaxRuleCollection;
use Shopware\Core\Checkout\Cart\Tax\TaxCalculator;
use Shopware\Core\Content\Category\Aggregate\CategoryTranslation\CategoryTranslationCollection;
use Shopware\Core\Content\Category\Aggregate\CategoryTranslation\CategoryTranslationEntity;
use Shopware\Core\Content\Category\CategoryCollection;
use Shopware\Core\Content\Media\MediaEntity;
use Shopware\Core\Content\Product\Aggregate\ProductManufacturer\ProductManufacturerEntity;
use Shopware\Core\Content\Product\Aggregate\ProductManufacturerTranslation\ProductManufacturerTranslationCollection;
use Shopware\Core\Content\Product\Aggregate\ProductManufacturerTranslation\ProductManufacturerTranslationEntity;
use Shopware\Core\Content\Product\Aggregate\ProductMedia\ProductMediaCollection;
use Shopware\Core\Content\Product\Aggregate\ProductMedia\ProductMediaEntity;
use Shopware\Core\Content\Product\Aggregate\ProductPrice\ProductPriceEntity;
use Shopware\Core\Content\Product\Aggregate\ProductTranslation\ProductTranslationEntity;
use Shopware\Core\Content\Product\ProductCollection;
use Shopware\Core\Content\Product\ProductEntity;
use Shopware\Core\Content\ProductStream\ProductStreamCollection;
use Shopware\Core\Content\Property\Aggregate\PropertyGroupOption\PropertyGroupOptionCollection;
use Shopware\Core\Content\Property\Aggregate\PropertyGroupOptionTranslation\PropertyGroupOptionTranslationCollection;
use Shopware\Core\Content\Property\Aggregate\PropertyGroupOptionTranslation\PropertyGroupOptionTranslationEntity;
use Shopware\Core\Content\Property\Aggregate\PropertyGroupTranslation\PropertyGroupTranslationCollection;
use Shopware\Core\Content\Property\Aggregate\PropertyGroupTranslation\PropertyGroupTranslationEntity;
use Shopware\Core\Content\Property\PropertyGroupEntity;
use Shopware\Core\Content\Seo\SeoUrlPlaceholderHandlerInterface;
use Shopware\Core\Framework\Context;
use Shopware\Core\Framework\DataAbstractionLayer\EntityCollection;
use Shopware\Core\Framework\DataAbstractionLayer\EntityRepository;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Criteria;
use Shopware\Core\Framework\DataAbstractionLayer\Search\EntitySearchResult;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\EqualsFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\MultiFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\NotFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Filter\RangeFilter;
use Shopware\Core\Framework\DataAbstractionLayer\Search\RequestCriteriaBuilder;
use Shopware\Core\Framework\DataAbstractionLayer\Pricing\Price;
use Shopware\Core\Framework\DataAbstractionLayer\Pricing\PriceCollection;
use Shopware\Core\Framework\DataAbstractionLayer\Search\Sorting\FieldSorting;
use Shopware\Core\Framework\Uuid\Uuid;
use Shopware\Core\System\Currency\CurrencyEntity;
use Shopware\Core\System\DeliveryTime\Aggregate\DeliveryTimeTranslation\DeliveryTimeTranslationEntity;
use Shopware\Core\System\DeliveryTime\DeliveryTimeEntity;
use Shopware\Core\System\SalesChannel\Aggregate\SalesChannelDomain\SalesChannelDomainCollection;
use Shopware\Core\System\SalesChannel\Aggregate\SalesChannelDomain\SalesChannelDomainEntity;
use Shopware\Core\System\SalesChannel\Context\SalesChannelContextFactory;
use Shopware\Core\System\SalesChannel\SalesChannelContext;
use Shopware\Core\System\SalesChannel\SalesChannelEntity;
use Shopware\Core\System\Tag\TagCollection;
use Shopware\Core\System\Tag\TagEntity;
use Shopware\Core\System\Tax\Aggregate\TaxRule\TaxRuleEntity;
use Shopware\Core\System\Tax\TaxEntity;
use Shopware\Core\System\Unit\UnitEntity;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route(defaults={"_routeScope"={"api"}})
 */
class ResChannableApiController extends AbstractController
{

    /**
     * @var Connection
     */
    private Connection $connection;

    /**
     * @var ResChannableConfigService
     */
    private $resChannableConfigService;

    /**
     * @var RequestCriteriaBuilder
     */
    private $requestCriteriaBuilder;

    /**
     * @var EntityRepository
     */
    private $productRepository;

    /**
     * @var EntityRepository
     */
    private $salesChannelRepository;

    /**
     * @var SalesChannelContextFactory
     */
    private $salesChannelContextFactory;

    /**
     * @var EntityRepository
     */
    private $currencyRepository;

    /**
     * @var SeoUrlPlaceholderHandlerInterface
     */
    private $seoUrlPlaceholderHandler;

    /**
     * @var EntityRepository
     */
    private $mediaRepository;

    /**
     * @var string
     */
    private $shopBaseUrl = '';

    /**
     * @var string
     */
    private $fallbackLanguageId = '';

    /**
     * @var array
     */
    private $priceRules = [];

    const DEFAULT_LIMIT = 250;

    /**
     * ResChannableApiController constructor.
     *
     * @param Connection $connection
     * @param ResChannableConfigService $resChannableConfigService
     * @param RequestCriteriaBuilder $requestCriteriaBuilder
     * @param EntityRepository $productRepository
     * @param EntityRepository $salesChannelRepository
     * @param SalesChannelContextFactory $salesChannelContextFactory
     * @param EntityRepository $currencyRepository
     * @param SeoUrlPlaceholderHandlerInterface $seoUrlPlaceholderHandler
     * @param EntityRepository $mediaRepository
     */
    public function __construct(
        Connection $connection,
        ResChannableConfigService $resChannableConfigService,
        RequestCriteriaBuilder $requestCriteriaBuilder,
        EntityRepository $productRepository,
        # No type declaration for reasons of compatibility
        $salesChannelRepository,
        # No type declaration for reasons of compatibility
        $salesChannelContextFactory,
        EntityRepository $currencyRepository,
        SeoUrlPlaceholderHandlerInterface $seoUrlPlaceholderHandler,
        # No type declaration for reasons of compatibility
        $mediaRepository
    )
    {
        $this->connection = $connection;
        $this->resChannableConfigService = $resChannableConfigService;
        $this->requestCriteriaBuilder = $requestCriteriaBuilder;
        $this->productRepository = $productRepository;
        $this->salesChannelRepository = $salesChannelRepository;
        $this->salesChannelContextFactory = $salesChannelContextFactory;
        $this->currencyRepository = $currencyRepository;
        $this->seoUrlPlaceholderHandler = $seoUrlPlaceholderHandler;
        $this->mediaRepository = $mediaRepository;
    }

    /**
     * Set webhook url
     *
     * @param Request $request
     * @param Context $context
     * @return JsonResponse
     * @Route("/api/_action/reschannable/setwebhookurl", name="api.action.reschannable.setwebhookurl", methods={"POST"})
     * @Route("/api/v{version}/_action/reschannable/setwebhookurl", name="api.action.reschannable.setwebhookurl.old", methods={"POST"})
     */
    public function setWebhookUrl(Request $request, Context $context): JsonResponse
    {
        $this->shopBaseUrl = $request ? $request->getSchemeAndHttpHost().$request->getBaseUrl() : '';

        $webhookResponse = $this->saveWebhookUrl($request, $context);

        $status = Response::HTTP_OK;
        if ( !$webhookResponse )
            $status = Response::HTTP_BAD_REQUEST;

        $jsonData = [
            'success' => $webhookResponse
        ];

        return new JsonResponse($jsonData,$status);
    }

    /**
     * Save webhook url
     *
     * @param Request $request
     * @param Context $context
     * @return bool
     */
    private function saveWebhookUrl(Request $request, Context $context): bool
    {
        $salesChannelId = $request->request->getAlnum('shop');
        $languageId = $request->request->getAlnum('lang');
        $webhookUrl = $request->request->get('url');

        if ( !$salesChannelId )
            return false;

        $webhookEnabled = true;
        if ( empty($webhookUrl) ) {
            $webhookUrl = '';
            $webhookEnabled = false;
        }

        # Looking for default language if lang id wasn't passed, e.g. for older connections
        if ( empty($languageId) ) {
            $criteria = new Criteria([$salesChannelId]);

            /** @var SalesChannelEntity|null $salesChannel */
            $salesChannel = $this->salesChannelRepository->search($criteria, $context)->get($salesChannelId);
            $languageId = $salesChannel->getLanguageId();
        }

        $this->resChannableConfigService->set('ResChannableConnector.settings.webhookUrl', $webhookUrl, $salesChannelId, $languageId);
        $this->resChannableConfigService->set('ResChannableConnector.settings.webhookEnabled', $webhookEnabled, $salesChannelId, $languageId);

        return true;
    }

    /**
     * Get articles
     *
     * @param Request $request
     * @param Context $context
     * @return JsonResponse
     * @Route("/api/_action/reschannable/getarticles", name="api.custom.reschannable.getarticles", methods={"GET"})
     * @Route("/api/v{version}/_action/reschannable/getarticles", name="api.custom.reschannable.getarticles.old", methods={"GET"})
     */
    public function getArticles(Request $request, Context $context): JsonResponse
    {
        $this->shopBaseUrl = $request ? $request->getSchemeAndHttpHost().$request->getBaseUrl() : '';
        $this->fallbackLanguageId = $context->getLanguageId();

        $articlesResponse = $this->getArticlesResponse($request, $context);
        $status = Response::HTTP_OK;

        if ( empty($articlesResponse) )
            $status = Response::HTTP_NO_CONTENT;

        return new JsonResponse($articlesResponse,$status);
    }

    /**
     * @param Request $request
     * @param Context $context
     * @return array
     */
    private function getArticlesResponse(Request $request, Context $context): array
    {
        $salesChannelId = $request->query->getAlnum('shop');
        $languageId = $request->query->getAlnum('lang');

        $criteria = new Criteria([$salesChannelId]);
        /** @var SalesChannelEntity|null $salesChannel */
        $salesChannel = $this->salesChannelRepository->search($criteria, $context)->get($salesChannelId);

        if ( !$salesChannel )
            return ['error' => 'Invalid sales channel','success' => 'false'];

        # Workaround for old connections if language is not passed
        if ( !$languageId )
            $languageId = $salesChannel->getLanguageId();

        $limit = (int) $this->resChannableConfigService->get('ResChannableConnector.settings.pollLimit',$salesChannelId, $languageId);

        if ( !$limit )
            $limit = self::DEFAULT_LIMIT;

        $offset = $request->query->getInt('offset');

        $this->priceRules = $this->resChannableConfigService->get('ResChannableConnector.settings.priceRules',$salesChannelId,$languageId);

        $response = [];
        $response['limit'] = $limit;
        $response['offset'] = $offset;
        $response['articles'] = $this->getArticleList($context, $salesChannel, $salesChannelId, $languageId, $limit, $offset);
        $response['success'] = ( empty($response['articles']) ? 'false' : 'true' );

        return $response;
    }

    /**
     * Get article list
     *
     * @param Context $context
     * @param SalesChannelEntity $salesChannel
     * @param string $salesChannelId
     * @param string $languageId
     * @param int $limit
     * @param int $offset
     *
     * @return array
     */
    private function getArticleList(Context $context, SalesChannelEntity $salesChannel, string $salesChannelId, string $languageId, $limit, $offset): array
    {
        $salesChannelContext = $this->createSalesChannelContext($salesChannel);
        $salesChannelContext->getContext()->setConsiderInheritance(true);
        $salesChannelContext->getContext()->assign(['languageIdChain' => [$languageId]]);

        $currencyId = $salesChannel->getCurrencyId();

        $criteria = new Criteria([$currencyId]);
        /** @var SalesChannelEntity|null $currency */
        $currency = $this->currencyRepository->search($criteria, $context)->get($currencyId);

        $countryId = $salesChannel->getCountryId();

        $articleElements = $this->getArticleElements($salesChannelId, $languageId, $context, $limit, $offset);

        $salesChannelDomain = $this->getSalesChannelDomain($salesChannelContext, $salesChannelId, $languageId);

        $productList = [];

        /** @var ProductEntity $element */
        foreach ( $articleElements as $element) {

            # Add item
            $productList[] = $this->getArticle($element, $languageId, $currency, $countryId, $context, $salesChannelContext, $salesChannelDomain);
        }

        return $productList;
    }

    /**
     * Get article
     *
     * @param ProductEntity $product
     * @param string $languageId
     * @param CurrencyEntity $currency
     * @param string $countryId
     * @param Context $context
     * @param SalesChannelContext $salesChannelContext
     * @param string $salesChannelDomain
     * @return array
     */
    private function getArticle(
        ProductEntity $product,
        string $languageId,
        CurrencyEntity $currency,
        string $countryId,
        Context $context,
        SalesChannelContext $salesChannelContext,
        string $salesChannelDomain): array
    {
        $parent = $this->getParentArticle($product, $context);

        return $this->parseArticle($product, $parent, $languageId, $currency, $countryId, $salesChannelContext, $salesChannelDomain);
    }

    /**
     * Get article entities
     *
     * @param string $salesChannelId
     * @param string $languageId
     * @param Context $context
     * @param int $limit
     * @param int $offset
     * @return array
     */
    private function getArticleElements(string $salesChannelId, string $languageId, Context $context, int $limit, int $offset): array
    {
        $exportActiveArticles = (int) $this->resChannableConfigService->get('ResChannableConnector.settings.activeArticlesEnabled',$salesChannelId,$languageId);
        $exportImageArticles = (int) $this->resChannableConfigService->get('ResChannableConnector.settings.onlyArticlesWithImgEnabled',$salesChannelId,$languageId);
        $exportEanArticles = (int) $this->resChannableConfigService->get('ResChannableConnector.settings.onlyArticlesWithEanEnabled',$salesChannelId,$languageId);
        $exportStockArticles = (int) $this->resChannableConfigService->get('ResChannableConnector.settings.onlyArticlesWithStockEnabled',$salesChannelId,$languageId);

        $criteria = new Criteria();
        $criteria->setLimit($limit);
        $criteria->setOffset($offset);

        # Sales channel assignment
        $criteria->addFilter(new MultiFilter(
            MultiFilter::CONNECTION_OR,
            [
                new EqualsFilter('product.visibilities.salesChannelId', $salesChannelId)
            ]
        ));

        # Only export active or linked products
        if ( $exportActiveArticles ) {

            $singleArticleFilter = new MultiFilter(
                MultiFilter::CONNECTION_AND,
                [
                    new EqualsFilter('product.active', true),
                    new EqualsFilter('product.parent.id', null),
                ]
            );

            $variantFilter = new MultiFilter(
                MultiFilter::CONNECTION_AND,
                [
                    new EqualsFilter('product.active', true),
                    new EqualsFilter('product.parent.active', true),
                ]
            );

            $inheritedVariantFilter = new MultiFilter(
                MultiFilter::CONNECTION_AND,
                [
                    new EqualsFilter('product.parent.active', true),
                    new EqualsFilter('product.active', null),
                ]
            );

            $criteria->addFilter(new MultiFilter(
                MultiFilter::CONNECTION_OR,
                [
                    $singleArticleFilter,
                    $variantFilter,
                    $inheritedVariantFilter
                ]
            ));
        }

        if ( $exportEanArticles ) {
            $criteria->addFilter(new NotFilter(
                NotFilter::CONNECTION_AND,
                [
                    new EqualsFilter('product.ean', null)
                ]
            ));
        }

        if ( $exportStockArticles ) {
            $criteria->addFilter(new MultiFilter(
                MultiFilter::CONNECTION_OR,
                [
                    new RangeFilter('product.availableStock', ['gt' => 0]),
                    new EqualsFilter('product.isCloseout', false)
                ]
            ));
        }

        if ( $exportImageArticles ) {
            $criteria->addFilter(new NotFilter(
                NotFilter::CONNECTION_AND,
                [
                    new EqualsFilter('product.cover.media.fileName', null)
                ]
            ));
        }

        $criteria->addAssociation('translations');
        $criteria->addAssociation('visibilities');
        $criteria->addAssociation('configuratorSettings.option');
        $criteria->addAssociation('manufacturer');
        $criteria->addAssociation('manufacturer.translations');
        $criteria->addAssociation('prices');
        $criteria->addAssociation('prices.rule');
        $criteria->addAssociation('tax');
        $criteria->addAssociation('tax.rules');
        $criteria->addAssociation('listingPrices');
        $criteria->addAssociation('searchKeywords');
        $criteria->addAssociation('unit');
        $criteria->addAssociation('deliveryTime');
        $criteria->addAssociation('deliveryTime.translations');
        $criteria->addAssociation('tags');

        # Media
        $criteria->addAssociation('cover');
        $criteria->addAssociation('media');

        # Custom fields
        $criteria->addAssociation('customFields');

        # Properties
        $criteria->addAssociation('properties');
        $criteria->addAssociation('properties.translations');
        $criteria->addAssociation('properties.group');
        $criteria->addAssociation('properties.group.translations');

        # Variants
        $criteria->addAssociation('options');
        $criteria->addAssociation('options.translations');
        $criteria->addAssociation('options.group');
        $criteria->addAssociation('options.group.translations');

        # Categories
        $criteria->addAssociation('categories');
        $criteria->addAssociation('categories.translations');
        $criteria->addAssociation('categoriesRo');
        $criteria->addAssociation('categoriesRo.translations');
        $criteria->addAssociation('mainCategories');

        # Dynamic product group details
        $criteria->addAssociation('streams');
        $criteria->addAssociation('streams.categories');
        $criteria->addAssociation('streams.categories.translations');

        # Export parent articles before variants
        $criteria->addSorting(new FieldSorting('parentId', FieldSorting::ASCENDING));

        # Then sort by article number to avoid duplicate articles
        $criteria->addSorting(new FieldSorting('productNumber', FieldSorting::ASCENDING));

        # Important for active flag inheritance, otherwise errors due to null value
        $context->setConsiderInheritance(true);

        /** @var ProductCollection $elements */
        $entity = $this->productRepository->search($criteria, $context);

        if (!$entity)
            return [];

        return $entity->getEntities()->getElements();
    }

    /**
     * Get parent article
     *
     * @param ProductEntity $product
     * @param Context $context
     * @return ProductEntity|null
     */
    public function getParentArticle(ProductEntity $product, Context $context): ?ProductEntity
    {
        $parent = null;
        $parentId = $product->getParentId();
        if ( $parentId ) {
            $criteria = new Criteria([$parentId]);
            $criteria->addAssociation('translations');
            $criteria->addAssociation('cover');
            $criteria->addAssociation('media');
            $criteria->addAssociation('manufacturer');
            $criteria->addAssociation('manufacturer.translations');
            $criteria->addAssociation('categories');
            $criteria->addAssociation('categories.translations');
            $criteria->addAssociation('categoriesRo');
            $criteria->addAssociation('categoriesRo.translations');
            $criteria->addAssociation('customFields');
            $criteria->addAssociation('mainCategories');
            $criteria->addAssociation('deliveryTime');
            $criteria->addAssociation('deliveryTime.translations');

            $entity = $this->productRepository->search($criteria, $context);

            if ( $entity instanceof EntitySearchResult )
                $parent = $entity->getEntities()->first();
        }

        return $parent;
    }

    /**
     * Parse article for export
     *
     * @param ProductEntity $product
     * @param ProductEntity|null $parent
     * @param string $languageId
     * @param CurrencyEntity $currency
     * @param string $countryId
     * @param SalesChannelContext $salesChannelContext
     * @param string $salesChannelDomain
     * @return array
     */
    public function parseArticle(
        ProductEntity $product,
        ?ProductEntity $parent,
        string $languageId,
        CurrencyEntity $currency,
        string $countryId,
        SalesChannelContext $salesChannelContext,
        string $salesChannelDomain): array
    {
        $translations = $product->getTranslations();
        $translation = $translations->filterByLanguageId($languageId)->first();

        # If no language is specified, use first language
        if ( !$translation instanceof ProductTranslationEntity )
            $translation = $translations->first();

        # Check parent translations
        $parentTranslation = null;
        if ( $parent instanceof ProductEntity ) {
            $parentTranslations = $parent->getTranslations();
            $parentTranslation = $parentTranslations->filterByLanguageId($languageId)->first();

            # If no language is specified, use first language
            if ( !$parentTranslation instanceof ProductTranslationEntity )
                $parentTranslation = $parentTranslations->first();
        }

        $article = [];
        $article = $this->parseGeneral($article, $product, $parent, $salesChannelContext, $translation, $parentTranslation, $languageId, $salesChannelDomain);
        $article = $this->parseImages($article, $product);
        $article = $this->parseCategories($article, $product, $parent, $languageId);
        $article = $this->parseOptions($article, $product, $languageId);
        $article = $this->parseProperties($article, $product, $languageId);
        $article = $this->parseCustomFields($article, $product, $parent, $translation, $parentTranslation, $salesChannelContext);
        $article = $this->parsePrice($article, $product, $currency, $countryId, $salesChannelContext);
        $article = $this->parseTags($article, $product);
        $article = $this->parseSearchKeywords($article, $product);

        return $this->parseManufacturer($article, $product, $languageId);
    }

    /**
     * Parse price for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param CurrencyEntity $currency
     * @param string $countryId
     * @param SalesChannelContext $salesChannelContext
     * @return array
     */
    private function parsePrice(array $article, ProductEntity $product, CurrencyEntity $currency, string $countryId, SalesChannelContext $salesChannelContext): array
    {
        # Price
        $priceCollection = $product->getPrice();

        $currencyId = $currency->getId();

        # No prices if no currency is present
        if ( !$currencyId ) {
            return $article;
        }

        if ( $priceCollection instanceof PriceCollection ) {

            $price = $priceCollection->getCurrencyPrice($currencyId);

            if ( $price instanceof Price ) {
                $priceGross = $price->getGross();
                $priceNet = $price->getNet();

                /** @var TaxEntity $tax */
                $tax = $salesChannelContext->getTaxRules()->filterByProperty('id', $product->getTax()->getId())->first();
                /** @var TaxRuleEntity $countryTaxRule */
                $countryTaxRule = $tax->getRules()->filterByProperty('countryId', $countryId)->first();

                if ($countryTaxRule instanceof TaxRuleEntity) {
                    $taxRules = new TaxRuleCollection([]);
                    $taxRules->add(
                        new TaxRule(
                            $countryTaxRule->getTaxRate()
                        )
                    );

                    $taxCalculator = new TaxCalculator();
                    /** @var CalculatedTax $calculatedTaxes */
                    $calculatedTaxes = $taxCalculator->calculateGrossTaxes($priceGross, $taxRules)->first();
                    $countryTax = $calculatedTaxes->getTax();

                    if ($countryTax) {
                        $priceNet = $priceGross - $countryTax;
                    }
                }

                $article['price'] = $priceGross;
                $article['priceNet'] = round($priceNet, 2);

                $listPrice = $price->getListPrice();
                if ($listPrice instanceof Price) {
                    $article['listPrice'] = $listPrice->getGross();
                    $article['listPriceNet'] = $listPrice->getNet();
                }

                # Shopware > 6.4
                if (method_exists($price, 'getRegulationPrice')) {
                    $cheapestPrice = $price->getRegulationPrice();
                    if ($cheapestPrice instanceof Price) {
                        $article['cheapestPrice'] = $cheapestPrice->getGross();
                        $article['cheapestPriceNet'] = $cheapestPrice->getNet();
                    }
                }
            }
        }

        $article['currency'] = $currency->getIsoCode();

        # Purchase price
        $purchasePricesCollection = $product->getPurchasePrices();
        if ( $purchasePricesCollection !== null ) {
            $purchasePrice = $purchasePricesCollection->getCurrencyPrice($currencyId);

            if ( $purchasePrice instanceof Price ) {
                $article['purchasePriceNet'] = $purchasePrice->getNet();
                $article['purchasePriceGross'] = $purchasePrice->getGross();
            }
        }

        # Price rules
        $prices = $product->getPrices();
        $priceRules = [];

        if ( $prices !== null ) {

            if ($this->priceRules) {
                $allowedPriceRules = $this->priceRules;

                $prices = $prices->filter(function(ProductPriceEntity $priceRule) use ($allowedPriceRules) {
                    return \in_array($priceRule->getRuleId(), $allowedPriceRules);
                });
            } else {
                $prices = $prices->getElements();
            }

            foreach ( $prices as $price ) {

                $rule = $price->getRule();
                $ruleName = $rule->getName();
                $rulePrice = $price->getPrice();

                if ( $rulePrice !== null ) {

                    $rulePrice = $rulePrice->getCurrencyPrice($currencyId);

                    if ( $rulePrice instanceof Price ) {

                        # Use ID if rule name empty
                        if ( !$ruleName )
                            $ruleName = $rule->getId();

                        # Field name for mapping
                        $ruleName = $this->filterFieldNames($ruleName) . '_' .
                            $price->getQuantityStart() . '_' .
                            $price->getQuantityEnd();

                        $priceRules[$ruleName] = [
                            'net' => $rulePrice->getNet(),
                            'gross' => $rulePrice->getGross(),
                        ];

                        $ruleListPrice = $rulePrice->getListPrice();
                        if ( $ruleListPrice instanceof Price ) {
                            $priceRules[$ruleName]['listPrice'] = $ruleListPrice->getGross();
                            $priceRules[$ruleName]['listPriceNet'] = $ruleListPrice->getNet();
                        }

                        # Shopware > 6.4
                        if ( method_exists($rulePrice,'getRegulationPrice')) {
                            $ruleCheapestPrice = $rulePrice->getRegulationPrice();
                            if ($ruleCheapestPrice instanceof Price) {
                                $priceRules[$ruleName]['cheapestPrice'] = $ruleCheapestPrice->getGross();
                                $priceRules[$ruleName]['cheapestPriceNet'] = $ruleCheapestPrice->getNet();
                            }
                        }
                    }
                }
            }
        }

        $article['priceRules'] = $priceRules;

        return $article;
    }

    /**
     * Parse manufacturer for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param string $languageId
     * @return array
     */
    private function parseManufacturer(array $article, ProductEntity $product, string $languageId): array
    {
        $manufacturer = $product->getManufacturer();

        if ( $manufacturer instanceof ProductManufacturerEntity ) {
            $manufacturerTranslations = $manufacturer->getTranslations();
            if ( $manufacturerTranslations instanceof ProductManufacturerTranslationCollection ) {
                $manufacturerTranslation = $manufacturerTranslations->filterByLanguageId($languageId)->first();
                if ( $manufacturerTranslation instanceof ProductManufacturerTranslationEntity ) {
                    $article['supplier'] = $manufacturerTranslation->getName();
                } else {
                    $manufacturerTranslation = $manufacturerTranslations->first();

                    if ( $manufacturerTranslation instanceof ProductManufacturerTranslationEntity ) {
                        $article['supplier'] = $manufacturerTranslation->getName();
                    }
                }
            }
        }

        return $article;
    }

    /**
     * Parse options for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param string $languageId
     * @return array
     */
    private function parseOptions(array $article, ProductEntity $product, string $languageId): array
    {
        $options = $product->getOptions();

        if ( $options instanceof PropertyGroupOptionCollection ) {

            /** @var PropertyGroupOptionCollection $options */
            $options = $options->getElements();
            $optionsSummarized = [];

            foreach ( $options as $option ) {

                $propertyGroupRaw = '';

                # Get property group translation
                $propertyGroupEntity = $option->getGroup();

                if ( $propertyGroupEntity instanceof PropertyGroupEntity ) {

                    $propertyGroupTranslation = $propertyGroupEntity->getTranslations();

                    if ( $propertyGroupTranslation instanceof PropertyGroupTranslationCollection ) {

                        $lngPropertyGroup = $propertyGroupTranslation->filterByLanguageId($languageId)->first();

                        if ( $lngPropertyGroup instanceof PropertyGroupTranslationEntity ) {
                            $propertyGroupRaw = $lngPropertyGroup->getName();
                            $propertyGroup = $this->filterFieldNames($propertyGroupRaw);
                        } else {
                            $lngPropertyGroup = $propertyGroupTranslation->first();

                            if ( $lngPropertyGroup instanceof PropertyGroupTranslationEntity ) {
                                $propertyGroupRaw = $lngPropertyGroup->getName();
                                $propertyGroup = $this->filterFieldNames($propertyGroupRaw);
                            } else {
                                $propertyGroupRaw = $propertyGroupEntity->getName();
                                $propertyGroup = $this->filterFieldNames($propertyGroupRaw);
                            }
                        }
                    }
                }

                # Get property translation
                $propertyTranslation = $option->getTranslations();

                if ( $propertyTranslation instanceof PropertyGroupOptionTranslationCollection ) {

                    # Connection lang
                    $lngProperty = $propertyTranslation->filterByLanguageId($languageId)->first();

                    $propertyName = '';

                    if ( $lngProperty instanceof PropertyGroupOptionTranslationEntity && $lngProperty->getName() ) {
                        $propertyName = $lngProperty->getName();
                    }

                    # Fallback lang
                    if ( !$propertyName ) {
                        $lngProperty = $propertyTranslation->filterByLanguageId($this->fallbackLanguageId)->first();

                        if ( $lngProperty instanceof PropertyGroupOptionTranslationEntity && $lngProperty->getName() ) {
                            $propertyName = $lngProperty->getName();
                        }
                    }

                    if ( $propertyName && $propertyGroup ) {
                        $article['options'][$propertyGroup] = $lngProperty->getName();
                        $optionsSummarized[] = $propertyGroupRaw . ": " . $lngProperty->getName();
                    }
                }
            }

            $article['optionsSummarized'] = implode(', ', $optionsSummarized);
        }

        return $article;
    }

    /**
     * Parse tags for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @return array
     */
    private function parseTags(array $article, ProductEntity $product): array
    {
        $tags = $product->getTags();

        if ( $tags instanceof TagCollection ) {

            /** @var TagCollection $tags */
            $tags = $tags->getElements();

            /** @var TagEntity $tag */
            foreach ( $tags as $tag ) {
                $article['tags'][] = $tag->getName();
            }
        }

        return $article;
    }

    /**
     * Parse search keywords for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @return array
     */
    private function parseSearchKeywords(array $article, ProductEntity $product): array
    {
        $searchKeywords = $product->getCustomSearchKeywords();

        if ( $searchKeywords )
            $article['searchKeywords'] = $searchKeywords;

        return $article;
    }

    /**
     * Parse properties for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param string $languageId
     * @return array
     */
    private function parseProperties(array $article, ProductEntity $product, string $languageId): array
    {
        $properties = $product->getProperties();

        if ( $properties instanceof PropertyGroupOptionCollection ) {

            /** @var PropertyGroupOptionCollection $properties */
            $properties = $properties->getElements();

            foreach ( $properties as $property ) {

                # Get property group translation
                $propertyGroupEntity = $property->getGroup();
                if ( $propertyGroupEntity instanceof PropertyGroupEntity ) {

                    $propertyGroupTranslation = $propertyGroupEntity->getTranslations();

                    if ( $propertyGroupTranslation instanceof PropertyGroupTranslationCollection ) {

                        $lngPropertyGroup = $propertyGroupTranslation->filterByLanguageId($languageId)->first();

                        if ( $lngPropertyGroup instanceof PropertyGroupTranslationEntity ) {
                            $propertyGroup = $this->filterFieldNames($lngPropertyGroup->getName());
                        } else {
                            $lngPropertyGroup = $propertyGroupTranslation->first();

                            if ( $lngPropertyGroup instanceof PropertyGroupTranslationEntity ) {
                                $propertyGroup = $this->filterFieldNames($lngPropertyGroup->getName());
                            } else {
                                $propertyGroup = $this->filterFieldNames($propertyGroupEntity->getName());
                            }
                        }
                    }
                }

                # Get property translation
                $propertyTranslation = $property->getTranslations();
                if ( $propertyTranslation instanceof PropertyGroupOptionTranslationCollection ) {

                    $lngProperty = $propertyTranslation->filterByLanguageId($languageId)->first();

                    if ( $lngProperty instanceof PropertyGroupOptionTranslationEntity ) {
                        $article['properties'][$propertyGroup][] = $lngProperty->getName();
                    } else {
                        $lngProperty = $propertyTranslation->first();
                        if ( $lngProperty instanceof PropertyGroupOptionTranslationEntity ) {
                            $article['properties'][$propertyGroup][] = $lngProperty->getName();
                        } else {
                            $article['properties'][$propertyGroup][] = $property->getName();
                        }
                    }
                }
            }
        }

        return $article;
    }

    /**
     * Parse custom fields for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param ProductEntity|null $parent
     * @param ProductTranslationEntity|null $translation
     * @param ProductTranslationEntity|null $parentTranslation
     * @param SalesChannelContext $salesChannelContext
     * @return array
     */
    private function parseCustomFields(
        array $article,
        ProductEntity $product,
        ?ProductEntity $parent,
        ?ProductTranslationEntity $translation,
        ?ProductTranslationEntity $parentTranslation,
        SalesChannelContext $salesChannelContext): array
    {
        $customFields = $product->getCustomFields();

        # Check parent custom fields
        if ($parent instanceof ProductEntity) {
            if (($parentCustomFields = $parent->getCustomFields()) !== null) {
                $customFields = array_filter($parentCustomFields, function($customContent): bool {
                    return $customContent !== null;
                });
            }
        }

        # Check parent translation
        if ($parentTranslation instanceof ProductTranslationEntity) {
            if (($parentLngCustomFields = $parentTranslation->getCustomFields()) !== null) {
                $customFields = array_merge($customFields, array_filter($parentLngCustomFields, function($customContent): bool {
                    return $customContent !== null;
                }));
            }
        }

        # Check variant translation
        if ($translation instanceof ProductTranslationEntity && ($translationCustomFields = $translation->getCustomFields()) !== null) {
            $customFields = array_merge($customFields, array_filter($translationCustomFields, function($customContent): bool {
                return $customContent !== null;
            }));
        }

        if ( is_array($customFields) ) {
            foreach ($customFields as $field => $values) {
                $article['attributes'][$this->filterFieldNames($field)] = $values;

                # Check media file
                if ( is_scalar($values) && Uuid::isValid((string) $values) ) {
                    /** @var MediaEntity $media */
                    $media = $this->mediaRepository->search(new Criteria([$values]), $salesChannelContext->getContext())->first();

                    if ( $media instanceof MediaEntity )
                        $article['attributes'][$this->filterFieldNames($field)] = $media->getUrl();
                }
            }
        }

        return $article;
    }

    /**
     * Parse categories for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param ProductEntity|null $parent
     * @param string $languageId
     * @return array
     */
    private function parseCategories(array $article, ProductEntity $product, ?ProductEntity $parent, string $languageId): array
    {
        $categories = $product->getCategories();

        if ( $categories instanceof CategoryCollection ) {

            $categories = $categories->getElements();
            $isParent = $article["isParent"];

            # Categories from dynamic product groups
            if ( method_exists($product,'getStreams') ) {
                $streams = $product->getStreams();

                if ($streams instanceof ProductStreamCollection) {
                    $streams = $streams->getElements();

                    foreach ($streams as $stream) {
                        $streamCategories = $stream->getCategories();

                        if ($streamCategories instanceof CategoryCollection) {
                            $streamCategories = $streamCategories->getElements();
                            $categories = $categories + $streamCategories;
                        }
                    }
                }
            }

            foreach ($categories as $category) {
                $categoryTranslation = $category->getTranslations();

                if ($categoryTranslation instanceof CategoryTranslationCollection) {
                    $lngCategory = $categoryTranslation->filterByLanguageId($languageId)->first();

                    if (!$lngCategory instanceof CategoryTranslationEntity)
                        $lngCategory = $categoryTranslation->first();

                    $lngBreadCrumb = $lngCategory->getBreadcrumb();

                    $path = $category->getPath();

                    if ($lngBreadCrumb && !empty($path)) {
                        $path = explode('|', trim($path, '|'));
                        $path[] = $category->getId();

                        $parsedBreadCrumb = [];
                        for ($i = 0; $i < count($path); $i++) {
                            $parsedBreadCrumb[] = $lngBreadCrumb[$path[$i]];
                        }

                        if ( $category->getProductStreamId() ) {
                            $article['dynamicCategories'][] = $parsedBreadCrumb;
                        }

                        $article['categories'][] = $parsedBreadCrumb;

                        if ($product->getMainCategories()->first() && $product->getMainCategories()->first()->getCategoryId() == $category->getId()) {
                            $article['seoCategory'] = $parsedBreadCrumb;
                        } elseif (!$isParent && $parent && ($parent->getMainCategories()->first() && $parent->getMainCategories()->first()->getCategoryId() == $category->getId())) {
                            $article['seoCategory'] = $parsedBreadCrumb;
                        }
                    }
                }
            }
        }

        return $article;
    }

    /**
     * Parse images for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @return array
     */
    private function parseImages(array $article, ProductEntity $product): array
    {
        # Set cover first
        $cover = $product->getCover();

        if ( $cover instanceof ProductMediaEntity ) {
            $cover = $cover->getMedia();
            if ( $cover )
                $article['images'][] = $cover->getUrl();
        }

        # Set images
        $mediaCollection = $product->getMedia();

        if ( $mediaCollection instanceof ProductMediaCollection ) {
            $mediaElements = $mediaCollection->getElements();

            foreach ( $mediaElements as $mediaEntity ) {
                $media = $mediaEntity->getMedia();

                if ( $media ) {
                    $url = $media->getUrl();

                    if ( !in_array($url,$article['images']) )
                        $article['images'][] = $url;
                }
            }
        }

        return $article;
    }

    /**
     * Parse general data for export
     *
     * @param array $article
     * @param ProductEntity $product
     * @param ProductEntity|null $parent
     * @param SalesChannelContext $salesChannelContext
     * @param ProductTranslationEntity|null $translation
     * @param ProductTranslationEntity|null $parentTranslation
     * @param string $languageId
     * @param string $salesChannelDomain
     * @return array
     */
    private function parseGeneral(
        array $article,
        ProductEntity $product,
        ?ProductEntity $parent,
        SalesChannelContext $salesChannelContext,
        ?ProductTranslationEntity $translation,
        ?ProductTranslationEntity $parentTranslation,
        string $languageId,
        string $salesChannelDomain): array
    {
        $releaseDate = $product->getReleaseDate();
        $createdAt = $product->getCreatedAt();
        $updatedAt = $product->getUpdatedAt();

        $unit = $product->getUnit();
        $purchaseUnit = $product->getPurchaseUnit();
        $referenceUnit = $product->getReferenceUnit();
        $packUnit = $product->getPackUnit();
        $tax = $product->getTax();
        $deliveryTime = null;

        # Check parent translations
        if ( $parent ) {
            if ( $parentName = $parent->getName() )
                $name = $parentName;
            if ( $parentDescriptionLong = $parent->getDescription() )
                $descriptionLong = $parentDescriptionLong;
            if ( $parentMetaTitle = $parent->getMetaTitle() )
                $metaTitle = $parentMetaTitle;
            if ( $parentMetaDesc = $parent->getMetaDescription() )
                $metaDescription = $parentMetaDesc;
            if ( $parentMetaKeywords = $parent->getKeywords() )
                $metaKeywords = $parentMetaKeywords;
            if ( ($parentDeliveryTime = $parent->getDeliveryTime()) instanceof DeliveryTimeEntity )
                $deliveryTime = $parentDeliveryTime;
            if ( $parentRestockTime = $parent->getRestockTime() )
                $restockTime = $parentRestockTime;
        }

        if ( $parentTranslation ) {
            if ( $parentName = $parentTranslation->getName() )
                $name = $parentName;
            if ( $parentLngDescription = $parentTranslation->getDescription() )
                $descriptionLong = $parentLngDescription;
            if ( $lngMetaTitle = $parentTranslation->getMetaTitle() )
                $metaTitle = $lngMetaTitle;
            if ( $lngMetaDesc = $parentTranslation->getMetaDescription() )
                $metaDescription = $lngMetaDesc;
            if ( $lngMetaKeywords = $parentTranslation->getKeywords() )
                $metaKeywords = $lngMetaKeywords;
        }

        # Check variant translations
        if ( $translation ) {
            if ( $lngName = $translation->getName() )
                $name = $lngName;
            if ( $lngDescriptionLong = $translation->getDescription() )
                $descriptionLong = $lngDescriptionLong;
            if ( $lngMetaTitle = $translation->getMetaTitle() )
                $metaTitle = $lngMetaTitle;
            if ( $lngMetaDesc = $translation->getMetaDescription() )
                $metaDescription = $lngMetaDesc;
            if ( $lngMetaKeywords = $translation->getKeywords() )
                $metaKeywords = $lngMetaKeywords;
        }

        if ( empty($name) )
            $name = $product->getName();
        if ( empty($descriptionLong) )
            $descriptionLong = $product->getDescription();
        if ( !$deliveryTime instanceof DeliveryTimeEntity )
            $deliveryTime = $product->getDeliveryTime();
        if ( empty($restockTime) )
            $restockTime = $product->getRestockTime();

        # Metas
        if ( empty($metaTitle) )
            $metaTitle = $product->getMetaTitle();
        if ( empty($metaDescription) )
            $metaDescription = $product->getMetaDescription();
        if ( empty($metaKeywords) )
            $metaKeywords = $product->getKeywords();

        # Delivery date
        $deliveryEarliest = '';
        $deliveryLatest = '';
        $deliveryMin = '';
        $deliveryMax = '';
        $deliveryUnit = '';
        $deliveryName = '';

        if ( $deliveryTime instanceof DeliveryTimeEntity ) {

            $deliveryTimeStruct = DeliveryTime::createFromEntity($deliveryTime);

            if ( $deliveryTimeStruct instanceof DeliveryTime ) {
                try {
                    $deliveryDate = DeliveryDate::createFromDeliveryTime($deliveryTimeStruct);

                    # Add restock time if available stock < 1
                    if ( $product->getAvailableStock() < 1 )
                        $deliveryDate = $deliveryDate->add(new \DateInterval('P' . $restockTime . 'D'));

                    $deliveryEarliest = $deliveryDate->getEarliest()->format('Y-m-d');
                    $deliveryLatest = $deliveryDate->getLatest()->format('Y-m-d');
                } catch ( Exception $exception )
                {
                }
            }

            $deliveryMin = $deliveryTime->getMin();
            $deliveryMax = $deliveryTime->getMax();
            $deliveryUnit = $deliveryTime->getUnit();
            $deliveryName = $deliveryTime->getName();

            $deliveryTimeTranslations = $deliveryTime->getTranslations();

            if ( $deliveryTimeTranslations instanceof EntityCollection) {
                $deliveryTimeTranslations = $deliveryTimeTranslations->filter(
                    function (DeliveryTimeTranslationEntity $deliveryTimeTranslation) use ($languageId) {
                        return $deliveryTimeTranslation->getLanguageId() === $languageId;
                    }
                );

                if ( $deliveryTimeTranslations instanceof EntityCollection ) {
                    $deliveryTimeTranslation = $deliveryTimeTranslations->first();

                    if ( $deliveryTimeTranslation instanceof DeliveryTimeTranslationEntity )
                        $deliveryName = $deliveryTimeTranslation->getName();
                }
            }
        }

        $url = $this->getSeoUrl(
            'frontend.detail.page',
            ['productId' => $product->getId()]
        );

        $url = $this->seoUrlPlaceholderHandler->replace($url, $salesChannelDomain, $salesChannelContext);

        $article['id'] = $product->getId();
        $article['parentId'] = $product->getParentId();
        $article['isParent'] = ( $product->getChildCount() > 0 );
        $article['articleNumber'] = $product->getProductNumber();
        $article['active'] = $product->getActive();
        $article['name'] = $name;
        $article['ean'] = $product->getEan();
        $article['descriptionLong'] = $descriptionLong;
        $article['releaseDate'] = ( $releaseDate instanceof \DateTimeInterface ? $releaseDate->format('Y-m-d') : '' );
        $article['images'] = [];
        $article['url'] = $url;
        $article['stock'] = $product->getAvailableStock();
        $article['minOrderQuantity'] = $product->getMinPurchase();
        $article['maxOrderQuantity'] = $product->getMaxPurchase();
        $article['purchaseSteps'] = $product->getPurchaseSteps();
        $article['price'] = 0;
        $article['priceNet'] = 0;
        $article['listPrice'] = null;
        $article['listPriceNet'] = null;
        $article['cheapestPrice'] = null;
        $article['cheapestPriceNet'] = null;
        $article['purchasePriceNet'] = null;
        $article['purchasePriceGross'] = null;
        $article['taxRate'] = ( $tax instanceof TaxEntity ? $tax->getTaxRate() : '' );
        $article['deliveryDateName'] = $deliveryName;
        $article['deliveryDateMin'] = $deliveryMin;
        $article['deliveryDateMax'] = $deliveryMax;
        $article['deliveryDateUnit'] = $deliveryUnit;
        $article['deliveryDateEarliest'] = $deliveryEarliest;
        $article['deliveryDateLatest'] = $deliveryLatest;
        $article['restockTime'] = $restockTime;
        $article['isCloseout'] = $product->getIsCloseout();
        $article['shippingFree'] = $product->getShippingFree();
        $article['weight'] = $product->getWeight();
        $article['width'] = $product->getWidth();
        $article['height'] = $product->getHeight();
        $article['length'] = $product->getLength();
        $article['supplier'] = '';
        $article['supplierNumber'] = $product->getManufacturerNumber();
        $article['categories'] = [];
        $article['dynamicCategories'] = [];
        $article['seoCategory'] = [];
        $article['properties'] = [];
        $article['attributes'] = [];
        $article['tags'] = [];
        $article['searchKeywords'] = [];
        $article['options'] = [];
        $article['optionsSummarized'] = '';
        $article['packUnit'] = $packUnit;
        $article['purchaseUnit'] = $purchaseUnit;
        $article['referenceUnit'] = $referenceUnit;
        $article['unit'] = ( $unit instanceof UnitEntity ? $unit->getShortCode() : '' );
        $article['unitName'] = ( $unit instanceof UnitEntity ? $unit->getName() : '' );
        $article['created'] = ( $createdAt instanceof \DateTimeInterface ? $createdAt->format('Y-m-d') : '' );
        $article['updated'] = ( $updatedAt instanceof \DateTimeInterface ? $updatedAt->format('Y-m-d') : '' );

        # Metas
        $article['metaTitle'] = $metaTitle;
        $article['metaDescription'] = $metaDescription;
        $article['metaKeywords'] = $metaKeywords;

        # Parent
        if ( $parent ) {
            $article['parentArticleNumber'] = $parent->getProductNumber();
            $article['parentName'] = $parentName;
        } else {
            $article['parentArticleNumber'] = '';
            $article['parentName'] = '';
        }

        return $article;
    }

    /**
     * Get SEO url
     *
     * @param $page
     * @param array $parameters
     * @return string
     */
    public function getSeoUrl($page, $parameters = []): string
    {
        return $this->seoUrlPlaceholderHandler->generate($page, $parameters);
    }

    /**
     * Create sales channel context
     *
     * @param SalesChannelEntity $salesChannel
     * @return SalesChannelContext
     */
    private function createSalesChannelContext(SalesChannelEntity $salesChannel): SalesChannelContext
    {
        return $this->salesChannelContextFactory->create(Uuid::randomHex(), $salesChannel->getId(), []);
    }

    /**
     * Remove bad chars from field names
     *
     * @param $field
     * @return string
     */
    private function filterFieldNames($field)
    {
        if ( !empty($field) ) {
            # Replace umlauts
            $field = str_replace(['Ä', 'Ö', 'Ü', 'ä', 'ö', 'ü', 'ß'], ['Ae', 'Oe', 'Ue', 'ae', 'oe', 'ue', 'ss'], $field);
            # Strip bad chars
            $field = preg_replace('/[^0-9a-zA-Z_]+/', '', $field);
            $field = strtolower($field);
        }
        return $field;
    }

    /**
     * Get domain url
     *
     * @param SalesChannelContext $salesChannelContext
     * @param string $salesChannelId
     * @param string $languageId
     * @return string
     */
    private function getSalesChannelDomain(SalesChannelContext $salesChannelContext, string $salesChannelId, string $languageId): string
    {
        $salesChannelDomainId = $this->resChannableConfigService->get('ResChannableConnector.settings.domain',$salesChannelId, $languageId);

        $domainCollection = $salesChannelContext->getSalesChannel()->getDomains();

        if ( !$domainCollection instanceof SalesChannelDomainCollection )
            return '';

        $domains = $domainCollection->filter(
            function (SalesChannelDomainEntity $salesChannelDomain) use ($salesChannelDomainId) {
                return $salesChannelDomain->getId() === $salesChannelDomainId;
            }
        );

        if ( !$domains instanceof SalesChannelDomainCollection )
            return '';

        $domain = $domains->first();

        if ( !$domain instanceof SalesChannelDomainEntity ) {

            # Use first domain found by language id
            $domains = $domainCollection->filter(
                function (SalesChannelDomainEntity $salesChannelDomain) use ($languageId) {
                    return $salesChannelDomain->getLanguageId() === $languageId;
                }
            );

            if ( !$domains instanceof SalesChannelDomainCollection )
                return '';

            $domain = $domains->first();
        }

        $url = '';
        if ( $domain instanceof SalesChannelDomainEntity )
            $url = trim($domain->getUrl(),'/');

        return $url;
    }

}
