<?php declare(strict_types=1);

namespace Res\ResChannableConnector\Setting;

use Doctrine\DBAL\Connection;
use Res\ResChannableConnector\Setting\Service\AutoConnectServiceInterface;
use Res\ResChannableConnector\Setting\Service\Config\ResChannableConfigService;
use Shopware\Core\Framework\Validation\DataBag\RequestDataBag;
use Symfony\Bundle\FrameworkBundle\Controller\AbstractController;
use Symfony\Component\HttpFoundation\JsonResponse;
use Symfony\Component\HttpFoundation\Request;
use Symfony\Component\HttpFoundation\Response;
use Symfony\Component\Routing\Annotation\Route;

/**
 * @Route(defaults={"_routeScope"={"api"}})
 */
class SettingsController extends AbstractController
{
    /**
     * @var ResChannableConfigService
     */
    private $resChannableConfigService;

    /**
     * @var AutoConnectServiceInterface
     */
    private $autoConnectService;

    /**
     * @var Connection
     */
    private $connection;

    public function __construct(ResChannableConfigService $resChannableConfigService, AutoConnectServiceInterface $apiService, Connection $connection)
    {
        $this->resChannableConfigService = $resChannableConfigService;
        $this->autoConnectService = $apiService;
        $this->connection = $connection;
    }

    /**
     * @Route("/api/_action/reschannable/get-auto-connect-url", name="api.action.reschannable.get.auto.connect.url", methods={"POST"})
     * @Route("/api/v{version}/_action/reschannable/get-auto-connect-url", name="api.action.reschannable.get.auto.connect.url.old", methods={"POST"})
     */
    public function getAutoConnectUrl(RequestDataBag $requestDataBag): JsonResponse
    {
        $salesChannelId = $requestDataBag->get('selectedSalesChannelId');
        $languageId = $requestDataBag->get('selectedLanguageId');

        $url = $this->autoConnectService->getUrl($salesChannelId, $languageId);

        return new JsonResponse(['url' => $url]);
    }

    /**
     * @Route("/api/_action/reschannable/get-config-values", name="api.action.reschannable.get.config.values", methods={"GET"})
     * @Route("/api/v{version}/_action/reschannable/get-config-values", name="api.action.reschannable.get.config.values.old", methods={"GET"})
     */
    public function getConfigValues(Request $request): JsonResponse
    {
        $languageId = $request->query->get('languageId');
        $salesChannelId = $request->query->get('salesChannelId');

        $config = $this->resChannableConfigService->getConfig($salesChannelId, $languageId);

        if (empty($config)) {
            $json = '{}';
        } else {
            $json = json_encode($config, JSON_PRESERVE_ZERO_FRACTION);
        }

        return new JsonResponse($json, 200, [], true);
    }

    /**
     * @Route("/api/_action/reschannable/save-config-values", name="api.action.reschannable.save.config.values", methods={"POST"})
     * @Route("/api/v{version}/_action/reschannable/save-config-values", name="api.action.reschannable.save.config.values.old", methods={"POST"})
     */
    public function saveConfigValues(Request $request): Response
    {
        $salesChannelId = $request->query->get('salesChannelId');
        $languageId = $request->query->get('languageId');
        $configs = $request->request->all();
        $this->saveKeyValues($salesChannelId, $languageId, $configs);

        return new Response('', Response::HTTP_NO_CONTENT);
    }
    /**
     * @Route("/api/_action/reschannable/batch-save-config-values", name="api.action.reschannable.batch.save.config.values", methods={"POST"})
     * @Route("/api/v{version}/_action/reschannable/batch-save-config-values", name="api.action.reschannable.batch.save.config.values.old", methods={"POST"})
     */
    public function batchSaveConfiguration(Request $request): Response
    {
        $this->connection->beginTransaction();

        foreach ($request->request->all() as $key => $config) {
            [$salesChannelId, $languageId] = explode('-', $key);

            $this->saveKeyValues($salesChannelId, $languageId, $config);
        }

        $this->connection->commit();

        return new Response('', Response::HTTP_NO_CONTENT);
    }

    private function saveKeyValues(string $salesChannelId, string $languageId, array $config): void
    {
        foreach ($config as $key => $value)
            $this->resChannableConfigService->set($key, $value, $salesChannelId, $languageId);
    }

}